@@ -2,7 +2,9 @@ from __future__ import absolute_import, print_function
 
 # TODO: rethink naming for 2D/3D functions
 
-from .geom2d import star_dist, polygons_to_label, relabel_image_stardist, ray_angles, dist_to_coord
+from .geom2d import star_dist, relabel_image_stardist, ray_angles, dist_to_coord, polygons_to_label, polygons_to_label_coord
 from .geom3d import star_dist3D, polyhedron_to_label, relabel_image_stardist3D, dist_to_coord3D, export_to_obj_file3D
 
+from .geom2d import _dist_to_coord_old, _polygons_to_label_old
+
 #, dist_to_volume, dist_to_centroid
